import bpy
import configparser
import sys
import urllib.request
import zipfile
from ..lua.loader import lua_loader
from bpy_extras.image_utils import load_image
from os import listdir
from os.path import isfile, exists, isfile
from mathutils import Matrix

_repos = [
	"https://ftp.train-fever.net/tools_archiv/blender_addon/repository.ini",
]

_plugin_events = [
	# import/export
	"init_import_operator",
	"draw_import_settings_panel",
	"init_export_operator",
	"draw_export_settings_panel",
	# model events
	"load_model_data",
	"validate_model_data",
	"load_lod",
	"load_model",
	"load_model_2", # for animation handling only for TPF2 loading
	"save_lod_data",
	"save_model_data",
	"save_model_data_2", # for animation handling only for TPF2 loading
	"create_ui_icon",
	"write_model_file",
	"draw_model_settings_panel",
	"draw_lod_settings_panel",
	# group events
	"load_group_data",
	"validate_group_data",
	"load_group",
	"save_group_data",
	"write_group_file",
	"draw_group_settings_panel",
	# mesh events
	"load_mesh_data",
	"load_mesh_data_bin",
	"validate_mesh_data",
	"validate_binary_mesh_data",
	"load_mesh_before_creation",
	"load_mesh_after_creation",
	"load_mesh_always",
	"get_mesh_vertex_data",
	"create_mesh_data_storage",
	"save_mesh_data_once",
	"save_mesh_data_always",
	"write_mesh_file",
	"draw_mesh_settings_panel",
	# material events
	"load_material_data",
	"validate_material_data",
	"load_material",
	"save_material_data",
	"write_material_file",
	"draw_material_settings_panel",
	"draw_texture_settings_panel",
]

# Path definiton for different object types
basepaths = [
  ['Train', 'vehicle/train/', ''],
  ['Waggon', 'vehicle/waggon/', ''],
  ['Bus', 'vehicle/bus/', ''],
  ['Tram', 'vehicle/tram/', ''],
  ['Car', 'vehicle/car/', ''],
  ['Truck', 'vehicle/truck/', ''],
  ['Ship', 'vehicle/ship/', ''],
  ['Plane', 'vehicle/plane/', ''],
  ['Station Bus', 'station/bus/', ''],
  ['Station Tram', 'station/tram/', ''],
  ['Station Truck', 'station/road/', ''],
  ['Station Train', 'station/train/', ''],
  ['Harbor', 'station/harbor/', ''],
  ['Airport', 'station/airport/', ''],
  ['Other', 'other', 'Use text field below']
]
	
class TF_base_functions():
	plugins = {
		"events": {event: [] for event in _plugin_events},
	}
	lua_scripts_basepaths = {}
	luaLoader = None
	import_settings = {}
	export_settings = {}
	subtype_list = [] #, ['Mod', 'mod', '', 6], ['Dependency', 'dependency', '', 7], ['Multiple Unit', 'multiple_unit', '', 8]]
	subtype_list_old = []
	export_file_ext = {}
	update = (False, "")
	addonID = "io_mesh_mdl"
	user_prefs = {
		"error_level": 0,
		"debug": False,
		"dds_converter": "",
		"update_auto_check": False,
		"update_branch": "stable",
	}
	user_settings = {
		"export_options": {
			"game": "TPF2",
			"hidden_faces": False,
			"unselected_faces": True,
			"ngons": True,
			"flipV": True,
			"override": False,
			"subfiles": True,
			"dir_structure": True,
			"material_files": True,
			"ui_icon": True,
			"bb": True,
			"mesh_directory": "",
			"model_type": "vehicle/train/",
			"model_type_other": "Insert path for material",
			"prefix": "",
			"ui_only": False,
			"textures": "TGA",
		},
		"import_options": {
			"lod_distance": 1,
			"mat_config": 0,
			"ui_only": False,
			"flat_faces": True,
			"game": "TPF2",
			"at_cursor": False,
		},
		"panels": {},
	}
	# user settings
	@classmethod
	def check_for_updates(cls):
		addon_version = getattr(sys.modules.get(bpy.context.user_preferences.addons[cls.addonID].module), "bl_info")["version"]

		repo_file = None
		for repo in _repos:
			try:
				repo_file = urllib.request.urlopen(repo)
			except:
				pass
			else:
				break
			
		if repo_file:
			repo_string = repo_file.read().decode()
			repo_data = configparser.ConfigParser()
			repo_data.read_string(repo_string)
			
			if cls.user_prefs["update_branch"] in repo_data:
				update_branch = repo_data[cls.user_prefs["update_branch"]]
				version = update_branch.get("version", "0.0.0").split(".")
				if len(version) == 3:
					website = update_branch.get("website", "")
					
					if int(version[0]) > addon_version[0] or (int(version[0]) == addon_version[0] and int(version[1]) > addon_version[1]) or (int(version[0]) == addon_version[0] and int(version[1]) == addon_version[1] and int(version[2]) > addon_version[2]):
						# new version available
						return (True, website)
		else:
			print("[io_mesh_mdl] could not connect to repository")
		
		return (False, "")
	
	@classmethod
	def read_user_settings(cls, filename):
		config = configparser.ConfigParser()
		files = config.read(filename)
		if len(files) > 0:
			for section in cls.user_settings.keys():
				if section in config:
					for key, string_value in config.items(section):
						value = string_value
						if key in cls.user_settings[section]:
							if isinstance(cls.user_settings[section][key], bool):
								value = config.getboolean(section, key)
							elif isinstance(cls.user_settings[section][key], int):
								value = int(value)
							elif isinstance(cls.user_settings[section][key], float):
								value = float(value)
						cls.user_settings[section][key] = value
		else:
			cls.save_user_settings(filename, config)
	
	@classmethod
	def save_user_settings(cls, filename, configParser = None):
		if not configParser:
			config = configparser.ConfigParser()
		else:
			config = configParser
		
		for section in sorted(cls.user_settings.keys()):
			config[section] = {}
			for key in sorted(cls.user_settings[section].keys()):
				config[section][key] = str(cls.user_settings[section][key])
		
		try:
			configfile = open(filename, "w")
		except:
			print("[io_mesh_mdl] Could not create user_settings.ini!")
		else:
			config.write(configfile)
			configfile.close()
	
	# Plugins
	@classmethod
	def plugin_add(cls, name, plugin):
		for event in _plugin_events:
			if hasattr(plugin, event):
				cls.plugins["events"][event].append({"function": getattr(plugin, event), "plugin": name})

	@classmethod
	def plugin_remove(cls, name):
		for event in cls.plugins["events"].values():
			i = 0
			while i < len(event):
				if event[i]["plugin"] == name:
					del event[i]
				else:
					i += 1

	@classmethod
	def call_plugin(cls, event, data, errors = []):
		if event in cls.plugins["events"]:
			for function in cls.plugins["events"][event]:
				plugin_reports = function["function"](data)
				if plugin_reports and errors:
					for report in plugin_reports:
						errors.append({"level": report["level"], "message": report["message"]})

	@classmethod
	def read_settings(cls):
		if cls.addonID in bpy.context.user_preferences.addons:
			if hasattr(bpy.context.user_preferences.addons[cls.addonID].preferences, "lua_scripts_basepath"):
				cls.lua_scripts_basepaths["TF"] = bpy.context.user_preferences.addons[cls.addonID].preferences.lua_scripts_basepath
			if hasattr(bpy.context.user_preferences.addons[cls.addonID].preferences, "lua_scripts_basepath_tpf"):
				cls.lua_scripts_basepaths["TPF"] = bpy.context.user_preferences.addons[cls.addonID].preferences.lua_scripts_basepath_tpf
			if hasattr(bpy.context.user_preferences.addons[cls.addonID].preferences, "lua_scripts_basepath_tpf2"):
				cls.lua_scripts_basepaths["TPF2"] = bpy.context.user_preferences.addons[cls.addonID].preferences.lua_scripts_basepath_tpf2
			if hasattr(bpy.context.user_preferences.addons[cls.addonID].preferences, "zip_extract_dir"):
				cls.import_settings["zip_extract_dir"] = bpy.context.user_preferences.addons[cls.addonID].preferences.zip_extract_dir
			if hasattr(bpy.context.user_preferences.addons[cls.addonID].preferences, "error_level"):
				cls.user_prefs["error_level"] = bpy.context.user_preferences.addons[cls.addonID].preferences.error_level
			if hasattr(bpy.context.user_preferences.addons[cls.addonID].preferences, "debug"):
				cls.user_prefs["debug"] = bpy.context.user_preferences.addons[cls.addonID].preferences.debug
			if hasattr(bpy.context.user_preferences.addons[cls.addonID].preferences, "dds_converter"):
				cls.user_prefs["dds_converter"] = bpy.context.user_preferences.addons[cls.addonID].preferences.dds_converter
			if hasattr(bpy.context.user_preferences.addons[cls.addonID].preferences, "update_auto_check"):
				cls.user_prefs["update_auto_check"] = bpy.context.user_preferences.addons[cls.addonID].preferences.update_auto_check
			if hasattr(bpy.context.user_preferences.addons[cls.addonID].preferences, "update_branch"):
				cls.user_prefs["update_branch"] = bpy.context.user_preferences.addons[cls.addonID].preferences.update_branch
		
		if cls.user_prefs["update_auto_check"]:
			cls.update = cls.check_for_updates()

	# Import functions
	@classmethod
	def create_4x4matrix(cls, values):
		output = Matrix()
		if len(values) >= 16:
			for i in range(0,4):
				for j in range(0,4):
					output[i][j] = float(values[4*j+i])
		return output

	@classmethod
	def split_path(cls, filename, errors = []):
		basepath = filename.rpartition("/res/")[0] + filename.rpartition("/res/")[1]
		filepath = filename.rpartition("/res/")[2]
		object_name = filepath.rpartition("/")[2]
		return basepath, filepath,object_name

	@classmethod
	def get_model_type(cls, filename, object_type, errors = []):
		modeltype = filename.rpartition("/res/")[2].partition(object_type)[2].rpartition("/")[0]
		if len(modeltype) > 0:
			modeltype += "/"
		return modeltype

	@classmethod
	def get_data(cls, filename, type, errors = []):
		if not cls.luaLoader:
			cls.luaLoader = lua_loader(cls.lua_scripts_basepaths[cls.import_settings["game"]], cls.import_settings["game"])

		default_data = {
			"MESH": {},
			"MESH_BIN": None,
			"GROUP": {"children": [], "matConfigs": []},
			"MODEL": {"lods": []},
			"MATERIAL": {"params": {}, "type": 'NONE'},
			"IMAGE": None,
			"MULTIPLE_UNIT": {"vehicles": {}},
			"ANIMATION": {"times": [], "transfs": []}
		}

		data = None

		if isfile(filename):
			if type == "MESH_BIN":
				data = open(filename, 'rb')
			elif type == "IMAGE":
				for image in bpy.data.images:
					if image.filepath.replace("\\", "/") == filename:
						data = image
						break
					else:
						errors.append({"level": 1, "message": "Images: " + image.filepath.replace("\\", "/") + " != " + filename})
				if not data:
					data = load_image(filename)
			else:
				data = cls.read_tf_script_file(filename, errors)
				if not data:
					data = default_data[type]
		else:
			errors.append({"level": 1, "message": "Could not find file: " + filename + ". Continue search in other directories."})
			new_filename = cls.find_tf_file(filename, errors)
			if new_filename == "":
				errors.append({"level": 2, "message": "Could not find file: " + filename + " in any directory."})
				data = default_data[type]
			else:
				if type == "MESH_BIN":
					data = open(new_filename, 'rb')
				elif type == "IMAGE":
					for image in bpy.data.images:
						if image.filepath.replace("\\", "/") == new_filename:
							data = image
							break
					if not data:
						data = load_image(new_filename)
				else:
					data = cls.read_tf_script_file(new_filename, errors)
					if not data:
						data = default_data[type]

		return data

	@classmethod
	def read_tf_script_file(cls, filename, errors):
		result = cls.luaLoader.load_lua_script(filename, True, errors)
		return result

	@classmethod
	def find_tf_file(cls, filepath, errors):
		tf_path = ""
		if "/mods/" in filepath:
			tf_path = filepath.rpartition("/mods/")[0]
			filename = filepath.rpartition("/mods/")[2]
			filename = filename.rpartition("/res/")[2]
		elif "/dlcs/" in filepath:
			tf_path = filepath.rpartition("/dlcs/")[0]
			filename = filepath.rpartition("/dlcs/")[2]
			filename = filepath.rpartition("/res/")[2]
		elif "/res/" in filepath:
			tf_path = filepath.rpartition("/res/")[0]
			filename = filepath.rpartition("/res/")[2]
		
		paths = []
		if tf_path != "":
			paths.append(tf_path) # The literal requested folder
		
		paths.append(cls.import_settings["basepath"]) # Seemingly the literal requested folder - maybe different if the 
		paths.append(cls.lua_scripts_basepaths[cls.import_settings["game"]]) # Should be the actual game folder
		
		for path in paths:
			# errors.append({"level": 1, "message": "Searching within directory '" + path + "'"})
			if isfile(path + "/res/" + filename):
				return path + "/res/" + filename
			else:
				# errors.append({"level": 1, "message": "Could not find file: " + path + "/res/" + filename + ". Continue search in other directories."})
				if exists(path + "/mods"):
					moddirs = listdir(path + "/mods")
					for moddir in moddirs:
						if isfile(path + "/mods/" + moddir + "/res/" + filename):
							return path + "/mods/" + moddir + "/res/" + filename
						#errors.append({"level": 1, "message": "Could not find file: " + path + "/mods/" + moddir + "/res/" + filename + ". Continue search in other directories."})
				
				if exists(path + "/dlcs"):
					dlcdirs = listdir(path + "/dlcs")
					for dlcdir in dlcdirs:
						if isfile(path + "/dlcs/" + dlcdir + "/res/" + filename):
							return path + "/dlcs/" + dlcdir + "/res/" + filename
						#errors.append({"level": 1, "message": "Could not find file: " + path + "/dlcs/" + dlcdir + "/res/" + filename + ". Continue search in other directories."})

		# In TPF2, files are zipped
		if cls.import_settings["game"] == "TPF2" and "models/" in filename:
			filename = filename.partition("models/")[2]
			if exists(cls.import_settings["zip_extract_dir"] + filename):
				return cls.import_settings["zip_extract_dir"] + filename
			else:
				zf = zipfile.ZipFile(cls.lua_scripts_basepaths[cls.import_settings["game"]]+"/res/models/models.zip")
				if filename in zf.namelist():
					errors.append({"level": 1, "message": "Extracting: " + filename + " to " + cls.import_settings["zip_extract_dir"] + ". Continue search in other directories."})
					zf.extract(filename, cls.import_settings["zip_extract_dir"])
					return cls.import_settings["zip_extract_dir"] + filename

		return ""

	# Export functions
	@classmethod
	def get_modeltype(cls, filepath, model_type, model_type_other):
		modeltype = (model_type, model_type_other)
		if filepath not in ("%auto", "%inherit"):
			modeltype = ("other", filepath)
			for i in range(len(basepaths)):
				if filepath == basepaths[i][1]:
					modeltype = (filepath, model_type_other)
					break
		
		return modeltype
	
	@classmethod
	def get_children(cls, object, addGroups=True):
		childList = []
		childs = []
		
		if hasattr(object, "children"):
			childs = object.children
		else:
			childs = object	
			
		for child in childs:
			if child.type == 'MESH':
				childList.append(child)
			elif child.type in ('EMPTY', 'ARMATURE'):
				if addGroups:
					childList.append(child)
				childList += cls.get_children(child, addGroups)
		
		return childList

	@classmethod
	def prop_to_value(cls, prop, errors=[]):
		value = None
		if str(type(prop)) == "<class 'bpy_prop_collection_idprop'>":
			value = []
			for item in prop:
				value.append(cls.prop_to_value(item))
		elif isinstance(prop, TFPropertyGroup):
			value = {}
			for sub_prop in prop.get_props():
				if isinstance(sub_prop, dict):
					if "name" in sub_prop:
						key = sub_prop["name"]
					else:
						key = sub_prop["prop_name"]
					prop_name = sub_prop["prop_name"]
				else:
					key = prop_name = sub_prop
				new_prop = getattr(prop, prop_name)
				if "index" in sub_prop:
					new_prop = new_prop[sub_prop["index"]]

				if hasattr(new_prop, "getValue"):
					item_value = new_prop.getValue()
				else:
					item_value = cls.prop_to_value(new_prop)
				if item_value != "" and item_value != {} and item_value != []:
					value[key] = item_value
		else:
			value = prop

		return value

	# UI functions
	@classmethod
	def is_hidden(cls, attr, default=True):
		isHidden = default
		if attr+"_hidden" in TF_base_functions.user_settings["panels"]:
			isHidden = TF_base_functions.user_settings["panels"][attr+"_hidden"]
			
			if TF_base_functions.user_settings["panels"][attr+"_hidden"] == default:
				del TF_base_functions.user_settings["panels"][attr+"_hidden"]

		return isHidden

	@classmethod
	def draw_hideable_box(cls, layout, name, id, default_hidden=True):
		box = None
		row = layout.row()
		row.label(name)
		if cls.is_hidden(id, default_hidden):
			row.operator("object.tfprop_show", icon="ZOOMIN", text="").attr = id
		else:
			row.operator("object.tfprop_hide", icon="ZOOMOUT", text="").attr = id
			box = layout.box()

		return box

	@classmethod
	def draw_internal_box(cls, data, layout, attr=""):
		boxIntern = cls.draw_hideable_box(layout, "Internal:", attr+"_internal")

		if boxIntern:
			if hasattr(data, "subtype"):
				boxIntern.prop(data, "subtype", text="Subtype")
			if hasattr(data, "model_type"):
				boxIntern.prop(data, "model_type", text="Model Type")
			if hasattr(data, "mesh_dir"):
				boxIntern.prop(data, "mesh_dir", text="Mesh Directory (%auto: use global setting)")

class TFPropertyGroup(bpy.types.PropertyGroup):
	prop_names = []
	@classmethod
	def get_props(cls):
		return cls.prop_names

	@classmethod
	def add_prop(cls, prop_name, prop, export=True):
		setattr(cls, prop_name, prop)
		if export and prop_name not in cls.prop_names:
			new_prop = prop_name
			if "name" in prop[1]:
				new_prop = {"name": prop[1]["name"], "prop_name": prop_name}
			cls.prop_names.append(new_prop)

	@classmethod
	def remove_prop(cls, prop_name):
		delattr(cls, prop_name)
		if prop_name in cls.prop_names:
			cls.prop_names.remove(prop_name)

class LuaExportFunctionProto():
	name = ""
	params = []
	def __init__(self, name="UnknownFunction", params=[]):
		self.name = name
		self.params = params