import bpy
import struct
from ..lua import writer
from . import material_util
from .base_functions import TF_base_functions
from mathutils import Vector

# Import
def is_valid_import_type(file_ext):
	return file_ext == "msh"

def _validate_data(data, errors):
	TF_base_functions.call_plugin("validate_mesh_data", data, errors)
	if str(type(data["meshData"])) == "<class 'dict'>":
		# check vertexData
		if "vertexAttr" in data["meshData"] and str(type(data["meshData"]["vertexAttr"])) == "<class 'dict'>":
			vertexData = data["meshData"]["vertexAttr"]
			# check for different position types
			if (not 'position' in vertexData) or (len(vertexData['position']) == 0):
				if ('positionIdx' in vertexData) and (len(vertexData['positionIdx']) > 0):
					vertexData['position'] = vertexData['positionIdx']
				elif ('positionAmbient' in vertex_data) and (len(vertexData['positionAmbient']) > 0):
					vertexData['position'] = vertexData['positionAmbient']

			# check for all required fields
			if not "position" in vertexData:
				errors.append({"level": 3, "message": "Missing field \"vertexAttr.position\" in Mesh \"%s\"!" % data["fileName"]})
				return False
			if not "normal" in vertexData:
				errors.append({"level": 3, "message": "Missing field \"vertexAttr.normal\" in Mesh \"%s\"!" % data["fileName"]})
				return False
			if not ("uv0" in vertexData or "uv1" in vertexData or "uv2" in vertexData):
				errors.append({"level": 3, "message": "Missing field \"vertexAttr.uv0\" in Mesh \"%s\"!" % data["fileName"]})
				return False

			# check if fields contain correct data
			for key in vertexData.keys():
				if str(type(vertexData[key])) != "<class 'dict'>":
					errors.append({"level": 3, "message": "Field \"vertexAttr.%s\" in Mesh \"%s\" has a wrong data format!" % (key, data["fileName"])})
					return False
				else:
					if not ("offset" in vertexData[key] and str(type(vertexData[key]["offset"])) in ("<class 'int'>", "<class 'long'>")):
						errors.append({"level": 3, "message": "Missing or wrong field \"vertexAttr.%s.offset\" in Mesh \"%s\"!" % (key, data["fileName"])})
						return False
					if not ("count" in vertexData[key] and str(type(vertexData[key]["count"])) in ("<class 'int'>", "<class 'long'>")):
						errors.append({"level": 3, "message": "Missing or wrong field \"vertexAttr.%s.count\" in Mesh \"%s\"!" % (key, data["fileName"])})
						return False
					if not ("numComp" in vertexData[key] and str(type(vertexData[key]["numComp"])) in ("<class 'int'>", "<class 'long'>")):
						errors.append({"level": 3, "message": "Missing or wrong field \"vertexAttr.%s.numComp\" in Mesh \"%s\"!" % (key, data["fileName"])})
						return False
		else:
			errors.append({"level": 3, "message": "Missing or wrong field \"vertexAttr\" in Mesh \"%s\"!" % data["fileName"]})
			return False

		# check subMeshes
		if "subMeshes" in data["meshData"] and str(type(data["meshData"]["subMeshes"])) == "<class 'list'>":
			if len(data["meshData"]["subMeshes"]) > 0:
				tris = 0
				for i in range(0, len(data["meshData"]["subMeshes"])):
					subMesh = data["meshData"]["subMeshes"][i]
					if str(type(subMesh)) == "<class 'dict'>":
						if "indices" in subMesh and str(type(subMesh["indices"])) == "<class 'dict'>":
							indexData = subMesh["indices"]
							# check for different position types
							if (not "position" in indexData) or str(type(indexData["position"])) != "<class 'dict'>" or len(indexData["position"]) == 0:
								if "positionIdx" in indexData and str(type(indexData["positionIdx"])) == "<class 'dict'>" and len(indexData["positionIdx"]) > 0:
									indexData["position"] = indexData["positionIdx"]
								elif 'positionAmbient' in indexData and str(type(indexData["positionAmbient"])) == "<class 'dict'>" and len(indexData["positionAmbient"]) > 0:
									indexData["position"] = indexData["positionAmbient"]

							# check for all required fields
							if not "position" in indexData:
								errors.append({"level": 3, "message": "Missing field \"subMeshes[%i].indices.position\" in Mesh \"%s\"!" % (i, data["fileName"])})
								return False
							if not "normal" in indexData:
								errors.append({"level": 3, "message": "Missing field \"subMeshes[%i].indices.normal\" in Mesh \"%s\"!" % (i, data["fileName"])})
								return False
							if not ("uv0" in indexData or "uv1" in indexData or "uv2" in indexData):
								errors.append({"level": 3, "message": "Missing field \"subMeshes[%i].indices.uv0\" in Mesh \"%s\"!" % (i, data["fileName"])})
								return False

							# check if fields contain correct data
							numTris = 0
							for key in indexData.keys():
								if str(type(indexData[key])) != "<class 'dict'>":
									errors.append({"level": 3, "message": "Field \"subMeshes[%i].indices.%s\" in Mesh \"%s\" has a wrong data format!" % (i, key, data["fileName"])})
									return False
								else:
									if not ("offset" in indexData[key] and str(type(indexData[key]["offset"])) in ("<class 'int'>", "<class 'long'>")):
										errors.append({"level": 3, "message": "Missing or wrong field \"subMeshes[%i].indices.%s.offset\" in Mesh \"%s\"!" % (i, key, data["fileName"])})
										return False
									if not ("count" in indexData[key] and str(type(indexData[key]["count"])) in ("<class 'int'>", "<class 'long'>")):
										errors.append({"level": 3, "message": "Missing or wrong field \"subMeshes[%i].indices.%s.count\" in Mesh \"%s\"!" % (i, key, data["fileName"])})
										return False
									if numTris > 0:
										if indexData[key]["count"] != numTris:
											errors.append({"level": 3, "message": "Different tri count for field \"subMeshes[%i].indices.%s.count\" in Mesh \"%s\"!" % (i, key, data["fileName"])})
											return False
										elif indexData[key]["count"] > 0:
											numTris = indexData[key]["count"]
											tris += indexData[key]["count"]
										else:
											errors.append({"level": 2, "message": "subMeshes %i in Mesh \"%s\" has no faces!" % (i, key, data["fileName"])})
									else:
										if indexData[key]["count"] > 0:
											numTris = indexData[key]["count"]
											tris += indexData[key]["count"]
										else:
											errors.append({"level": 2, "message": "subMeshes %i in Mesh \"%s\" has no faces!" % (i, key, data["fileName"])})

						else:
							errors.append({"level": 3, "message": "Missing or wrong field \"subMeshes[%i].indices\" in Mesh \"%s\"!" % (i, data["fileName"])})
							return False

						if TF_base_functions.import_settings["game"] != "TPF2":
							if "materials" not in subMesh or str(type(subMesh["materials"]))!= "<class 'list'>" or len(subMesh["materials"]) == 0:
								errors.append({"level": 3, "message": "Missing or wrong field \"subMeshes[%i].materials\" in Mesh \"%s\"!" % (i, data["fileName"])})
								return False
					else:
						errors.append({"level": 3, "message": "Field \"subMeshes[%i]\" in Mesh \"%s\" has a wrong data format!" % (i, data["fileName"])})
						return False
				if tris == 0:
					errors.append({"level": 3, "message": "Mesh \"%s\" has no faces!" % data["fileName"]})
					return False
			else:
				errors.append({"level": 3, "message": "Mesh \"%s\" has no subMesh!" % data["fileName"]})
				return False
		else:
			errors.append({"level": 3, "message": "Missing or wrong field \"subMeshes\" in Mesh \"%s\"!" % data["fileName"]})
			return False

		# check matConfig
		if TF_base_functions.import_settings["game"] != "TPF2":
			if "matConfigs" in data["meshData"] and str(type(data["meshData"]["matConfigs"])) == "<class 'list'>":
				if data["matConfig"] >= len(data["meshData"]["matConfigs"]):
					errors.append({"level": 2, "message": "Selected MatConfig index for Mesh \"%s\" is too high!" % data["fileName"]})
					data["matConfig"] = len(data["meshData"]["matConfigs"])-1
				matConfig = data["meshData"]["matConfigs"][data["matConfig"]]
				if str(type(matConfig)) == "<class 'list'>":
					if len(data["meshData"]["subMeshes"]) > len(matConfig):
						errors.append({"level": 2, "message": "MatConfig %i of mesh \"%s\" has not enough values!" % (data["matConfig"], data["fileName"])})
						for i in range(len(matConfig), len(data["meshData"]["subMeshes"])):
							matConfig.append(0)
					elif len(data["meshData"]["subMeshes"]) < len(matConfig):
						errors.append({"level": 1, "message": "MatConfig %i of mesh \"%s\" has too much values!" % (data["matConfig"], data["fileName"])})
				else:
					errors.append({"level": 3, "message": "Missing or wrong field \"matConfigs[%i]\" in Mesh \"%s\"!" % (data["matConfig"], data["fileName"])})
					return False
			else:
				errors.append({"level": 3, "message": "Missing or wrong field \"matConfigs\" in Mesh \"%s\"!" % data["fileName"]})
				return False
	else:
		errors.append({"level": 3, "message": "Mesh \"%s\" could not be loaded or has a wrong data format!" % data["fileName"]})
		return False

	return True

def _validate_binary_data(data, errors):
	TF_base_functions.call_plugin("validate_binary_mesh_data", data, errors)
	if not "meshDataBin" in data or not data["meshDataBin"]:
		errors.append({"level": 3, "message": "Could not read binary data for Mesh \"%s\"!" % data["fileName"]})
		return False
	else:
		vertexData = data["meshData"]["vertexAttr"]
		for key in vertexData.keys():
			if vertexData[key]["offset"] + vertexData[key]["count"] > data["meshDataBin"].seek(0, 2):
				errors.append({"level": 3, "message": "Unexpected <eof> in binary data for Mesh \"%s\"!" % data["fileName"]})
				data["meshDataBin"].close()
				return False
		subMeshes = data["meshData"]["subMeshes"]
		for subMesh in subMeshes:
			for key in subMesh["indices"].keys():
				if subMesh["indices"][key]["offset"] + subMesh["indices"][key]["count"] > data["meshDataBin"].seek(0, 2):
					errors.append({"level": 3, "message": "Unexpected <eof> in binary data for Mesh \"%s\"!" % data["fileName"]})
					data["meshDataBin"].close()
					return False

	return True

def _create_mesh(mesh_name, data, flipV, errors):
	tris = data["tris"]
	vertex_data = data["vertexData"]
	material_data = data["materialData"]
	
	if TF_base_functions.import_settings["game"] != "TPF2":
		matConfig = data["meshData"]["matConfigs"][data["matConfig"]]

	# count faces
	indexnum = 0
	for i in range(0, len(tris)):
		indexnum = indexnum + len(tris[i]["position"])

	# create mesh and add faces
	new_mesh = bpy.data.meshes.new(mesh_name)
	new_mesh.tessfaces.add(indexnum)

	# add UV maps
	uvmap = {}
	for i in range(0,10):
		if "uv"+str(i) in vertex_data:
			uvmap[i] = new_mesh.tessface_uv_textures.new()

	# var initialization
	index_start = 0
	vertex_start = 0
	vertexnum = 0
	
	faceToVertex = []

	for i in range(len(tris)):
		created_vertices = {}
		created_vertices_pos = {}
		indexnum = len(tris[i]["position"])
		faceToVertex.append([])
		new_verts = []
		material = None
		uvs = []
		for k in range(0,10):
			if "uv"+str(k) in vertex_data and k in uvmap:
				uvs.append((k, "uv"+str(k)))
		
		# get material for this submesh
		if i < len(material_data):
			if TF_base_functions.import_settings["game"] == "TPF2":
				material = material_data[i]
			else:
				material = material_data[i][matConfig[i]]
			new_mesh.materials.append(material)
		else:
			new_mesh.materials.append(None)
			
		# get texture for 3D view
		if material:
			material_diffuse_texture = material_util.get_diffuse_texture(material)
		else:
			material_diffuse_texture = None
		
		for j in range(indexnum):
			face_verts = [0, 0, 0]
			face_index = index_start + j
			is_flat = data["flat_faces"]
			vert_string_nor = tuple(vertex_data["normal"][tris[i]["normal"][j][0]])
			for k in range(1, 3):
				is_flat = is_flat and (tuple(vertex_data["normal"][tris[i]["normal"][j][k]]) == vert_string_nor)

			for k in range(3):
				vert_pos = Vector(vertex_data["position"][tris[i]["position"][j][k]])
				vert_string_pos = vert_pos.to_tuple(6)
				vert_normal = Vector(vertex_data["normal"][tris[i]["normal"][j][k]])
				vert_string = vert_string_pos + vert_normal.to_tuple(6)
				
				if is_flat and (vert_string_pos in created_vertices_pos):
					face_verts[k] = created_vertices_pos[vert_string_pos]
				elif vert_string in created_vertices: # not is_flat
					face_verts[k] = created_vertices[vert_string]
				else:
					new_verts.append((vert_pos, vert_normal))
					if is_flat:
						created_vertices_pos[vert_string_pos] = vertexnum
					else:
						created_vertices[vert_string] = vertexnum
					face_verts[k] = vertexnum
					vertexnum += 1
				
			faceToVertex[i].append(face_verts)

			# write vertex data to mesh
			new_mesh.tessfaces[face_index].vertices = face_verts
			new_mesh.tessfaces[face_index].use_smooth = not is_flat
			for uv in uvs:
				uvmap[uv[0]].data[face_index].uv1 = Vector(vertex_data[uv[1]][tris[i][uv[1]][j][0]])
				uvmap[uv[0]].data[face_index].uv2 = Vector(vertex_data[uv[1]][tris[i][uv[1]][j][1]])
				uvmap[uv[0]].data[face_index].uv3 = Vector(vertex_data[uv[1]][tris[i][uv[1]][j][2]])
				
				if material_diffuse_texture and (material_diffuse_texture.name[-3:] == "dds" or ".dds." in material_diffuse_texture.name): # DDS images are flipped because reasons
					uvmap[uv[0]].data[face_index].uv1.y = 1 - uvmap[uv[0]].data[face_index].uv1.y
					uvmap[uv[0]].data[face_index].uv2.y = 1 - uvmap[uv[0]].data[face_index].uv2.y
					uvmap[uv[0]].data[face_index].uv3.y = 1 - uvmap[uv[0]].data[face_index].uv3.y


		# add vertices to mesh
		new_mesh.vertices.add(len(new_verts))
		for j in range(len(new_verts)):
			new_mesh.vertices[vertex_start + j].co = new_verts[j][0]
			new_mesh.vertices[vertex_start + j].normal = new_verts[j][1]

		for j in range(index_start, index_start + indexnum):
			new_mesh.tessfaces[j].material_index = i
			if material_diffuse_texture:
				new_mesh.tessface_uv_textures[0].data[j].image = material_diffuse_texture

		# set new offsets
		index_start = index_start + indexnum
		vertex_start = vertex_start + len(new_verts)
	
	data["faceToVertex"] = faceToVertex

	# check mesh for errors
	new_mesh.update(False, True)
	new_mesh.validate()

	# create mesh object
	new_obj = bpy.data.objects.new(mesh_name, new_mesh)

	# link object to scene
	scene = bpy.context.scene
	scene.objects.link(new_obj)
	return new_obj

def load_from_file(filename, matConfig, select, errors):
	if (TF_base_functions.user_prefs["debug"]):
		print("\tload_mesh_file", filename)

	flat_faces = TF_base_functions.user_settings["import_options"]["flat_faces"]
	if "flat_faces" in TF_base_functions.import_settings:
		flat_faces = TF_base_functions.import_settings["flat_faces"]
	new_mesh = None

	basepath, filepath, mesh_name = TF_base_functions.split_path(filename)
	if basepath == "":
		errors.append({"level": 1, "message": "Could not find \"res\"-directory in \"%s\", loading of materials probably won't be succesfull." %filename})
		basepath = TF_base_functions.import_settings["basepath"] + "/res/"
	
	mesh_name = mesh_name.replace(".msh", "")
	modeltype = TF_base_functions.get_model_type(filename, "models/mesh/")

	# load mesh data from file
	mesh_data = TF_base_functions.get_data(basepath + filepath, "MESH", errors)
	data = {
		"meshData": mesh_data,
		"fileName": filename,
		"matConfig": matConfig,
		"name": mesh_name,
		"modelType": modeltype,
		"basepath": basepath,
	}
	TF_base_functions.call_plugin("load_mesh_data", data, errors)

	if _validate_data(data, errors):
		mesh_exists = False
		if bpy.data.meshes.find(mesh_name) > -1:
			if bpy.data.meshes[mesh_name].model_type == modeltype:
				mesh_exists = True
				new_mesh = bpy.data.meshes[mesh_name]
			else:
				for mesh in bpy.data.meshes:
					if mesh.name.find(mesh_name+".") == 0 and mesh.model_type == modeltype:
						mesh_exists = True
						new_mesh = mesh

		if mesh_exists:
			# mesh already exists
			# create mesh object
			new_obj = bpy.data.objects.new(mesh_name, new_mesh)
			scene = bpy.context.scene
			scene.objects.link(new_obj)

		else:
			# load mesh from files
			submeshes = mesh_data["subMeshes"]
			vertex = mesh_data["vertexAttr"]
			material_data = [[] for i in range(0, len(submeshes))]
			for i in range(0, len(submeshes)):
				for j in range(0, len(submeshes[i]["materials"])):
					mat_filename = submeshes[i]["materials"][j]
					material_data[i].append(material_util.load_from_file(basepath + "models/material/" + mat_filename, errors))

			# load binary mesh data from file
			tf_object = TF_base_functions.get_data(basepath + filepath + ".blob", "MESH_BIN", errors)
			data["meshDataBin"] = tf_object
			TF_base_functions.call_plugin("load_mesh_data_bin", data, errors)

			if _validate_binary_data(data, errors):
				vertex_data = {}
				vertex_data_num = {}
				for key in vertex.keys():
					vertex_data_num[key] = vertex[key]["count"]//(vertex[key]["numComp"]*4)
					vertex_data[key] = [[0.0 for i in range(0, vertex[key]["numComp"])] for j in range(0, vertex_data_num[key])]
					tf_object.seek(vertex[key]["offset"])
					for i in range(0, vertex_data_num[key]):
						vertex_data[key][i] = struct.unpack('<'+str(vertex[key]["numComp"])+'f', tf_object.read(4*vertex[key]["numComp"]))

				index_data = [{} for i in range(0, len(submeshes))]
				index_data_num = [{} for i in range(0, len(submeshes))]
				tris = [{} for i in range(0, len(submeshes))]
				for k in range(0, len(submeshes)):
					key_list = submeshes[k]["indices"].keys()
					for key in key_list:
						tris[k][key] = []
						index_data_num[k][key] = submeshes[k]["indices"][key]["count"]//4
						index_data[k][key] = [0 for j in range(0, index_data_num[k][key])]
						tf_object.seek(submeshes[k]["indices"][key]["offset"])
						for i in range(0, index_data_num[k][key]):
							 index_data[k][key][i] = struct.unpack("<L", tf_object.read(4))[0]
						tris[k][key] = [[0, 0, 0] for i in range(0, index_data_num[k][key]//3)]
						for i in range(0, len(tris[k][key])):
							tris[k][key][i] = [index_data[k][key][3*i], index_data[k][key][3*i+1], index_data[k][key][3*i+2]]

				tf_object.close()
				data["materialData"] = material_data
				data["tris"] = tris
				data["vertexData"] = vertex_data
				data["flat_faces"] = flat_faces

				TF_base_functions.call_plugin("load_mesh_before_creation", data, errors)

				# create mesh object and set properties
				new_obj = _create_mesh(mesh_name, data, False, errors)
				if new_obj:
					new_obj.data.model_type = modeltype
					new_obj.data.mesh_dir = ""
					new_mesh = new_obj.data

					data["meshObj"] = new_obj
					TF_base_functions.call_plugin("load_mesh_after_creation", data, errors)
			else:
				new_obj = None

		data["meshObj"] = new_obj
		if TF_base_functions.import_settings["at_cursor"]:
			new_obj.location = bpy.context.scene.cursor_location
		TF_base_functions.call_plugin("load_mesh_always", data, errors)

		if new_obj and new_mesh:
			new_mesh.update()
			new_obj.select = select

		return new_obj
	return None

def load_from_file_TPF2(filename, name, matlist, select, errors):
	if (TF_base_functions.user_prefs["debug"]):
		print("\tload_mesh_file", filename)

	flat_faces = TF_base_functions.user_settings["import_options"]["flat_faces"]
	if "flat_faces" in TF_base_functions.import_settings:
		flat_faces = TF_base_functions.import_settings["flat_faces"]
	new_mesh = None

	basepath, filepath, mesh_name = TF_base_functions.split_path(filename)
	if basepath == "":
		errors.append({"level": 1, "message": "Could not find \"res\"-directory in \"%s\", loading of materials probably won't be succesful." %filename})
		basepath = TF_base_functions.import_settings["basepath"] + "/res/"
	
	mesh_name = mesh_name.replace(".msh", "")
	mesh_filename = mesh_name
	if name != "":
		mesh_name = name
	modeltype = TF_base_functions.get_model_type(filename, "models/mesh/")

	# load mesh data from file
	mesh_data = TF_base_functions.get_data(basepath + filepath, "MESH", errors)
	data = {
		"meshData": mesh_data,
		"fileName": filename,
		"name": mesh_name,
		"modelType": modeltype,
		"basepath": basepath,
	}
	TF_base_functions.call_plugin("load_mesh_data", data, errors)

	if _validate_data(data, errors):
		mesh_exists = False
		if bpy.data.meshes.find(mesh_name) > -1:
			if bpy.data.meshes[mesh_name].model_type == modeltype:
				mesh_exists = True
				new_mesh = bpy.data.meshes[mesh_name]
			else:
				for mesh in bpy.data.meshes:
					if mesh.name.find(mesh_name+".") == 0 and mesh.model_type == modeltype:
						mesh_exists = True
						new_mesh = mesh

		if mesh_exists:
			# mesh already exists
			# create mesh object
			new_obj = bpy.data.objects.new(mesh_name, new_mesh)
			scene = bpy.context.scene
			scene.objects.link(new_obj)

		else:
			# load mesh from files
			submeshes = mesh_data["subMeshes"]
			vertex = mesh_data["vertexAttr"]
			material_data = []
			for i in range(0, len(submeshes)):
				mat_filename = matlist[i]
				material_data.append(material_util.load_from_file(basepath + "models/material/" + mat_filename, errors))

			# load binary mesh data from file
			tf_object = TF_base_functions.get_data(basepath + filepath + ".blob", "MESH_BIN", errors)
			data["meshDataBin"] = tf_object
			TF_base_functions.call_plugin("load_mesh_data_bin", data, errors)

			if _validate_binary_data(data, errors):
				vertex_data = {}
				vertex_data_num = {}
				for key in vertex.keys():
					vertex_data_num[key] = vertex[key]["count"]//(vertex[key]["numComp"]*4)
					vertex_data[key] = [[0.0 for i in range(0, vertex[key]["numComp"])] for j in range(0, vertex_data_num[key])]
					tf_object.seek(vertex[key]["offset"])
					for i in range(0, vertex_data_num[key]):
						vertex_data[key][i] = struct.unpack('<'+str(vertex[key]["numComp"])+'f', tf_object.read(4*vertex[key]["numComp"]))

				index_data = [{} for i in range(0, len(submeshes))]
				index_data_num = [{} for i in range(0, len(submeshes))]
				tris = [{} for i in range(0, len(submeshes))]
				for k in range(0, len(submeshes)):
					key_list = submeshes[k]["indices"].keys()
					for key in key_list:
						tris[k][key] = []
						index_data_num[k][key] = submeshes[k]["indices"][key]["count"]//4
						index_data[k][key] = [0 for j in range(0, index_data_num[k][key])]
						tf_object.seek(submeshes[k]["indices"][key]["offset"])
						for i in range(0, index_data_num[k][key]):
							 index_data[k][key][i] = struct.unpack("<L", tf_object.read(4))[0]
						tris[k][key] = [[0, 0, 0] for i in range(0, index_data_num[k][key]//3)]
						for i in range(0, len(tris[k][key])):
							tris[k][key][i] = [index_data[k][key][3*i], index_data[k][key][3*i+1], index_data[k][key][3*i+2]]

				tf_object.close()
				data["materialData"] = material_data
				data["tris"] = tris
				data["vertexData"] = vertex_data
				data["flat_faces"] = flat_faces

				TF_base_functions.call_plugin("load_mesh_before_creation", data, errors)

				# create mesh object and set properties
				new_obj = _create_mesh(mesh_name, data, True, errors)
				if new_obj:
					new_obj.data.model_type = modeltype
					new_obj.data.filename = mesh_filename
					new_obj.data.mesh_dir = ""
					new_mesh = new_obj.data

					data["meshObj"] = new_obj
					TF_base_functions.call_plugin("load_mesh_after_creation", data, errors)
			else:
				new_obj = None

		data["meshObj"] = new_obj
		if TF_base_functions.import_settings["at_cursor"]:
			new_obj.location = bpy.context.scene.cursor_location
		TF_base_functions.call_plugin("load_mesh_always", data, errors)

		if new_obj and new_mesh:
			new_mesh.update()
			new_obj.select = select

		return new_obj
	return None

# Export
def is_valid_object(object):
	return object.type == 'MESH'

def get_export_type(object):
	return "msh"

def get_model_type(object, model_type, model_type_other):
	return TF_base_functions.get_modeltype(object.data.model_type, model_type, model_type_other)

def get_mesh_dir(object, mesh_dir):
	if object.data.mesh_dir != "%auto":
		return object.data.mesh_dir
	else:
		return mesh_dir

def get_filename(object):
	return object.data.name

def create_file_data(mesh_object, name, errors):
	mesh_obj = mesh_object["object"]

	# get name
	mesh_name = name
	if mesh_name == "":
		mesh_name =  TF_base_functions.export_settings["prefix"] + mesh_obj.data.filename # this is the thing that actually changes the name of the output file

	if (TF_base_functions.user_prefs["debug"]):
		print("\tsave_mesh_file", TF_base_functions.export_settings["prefix"] + mesh_obj.data.filename)

	if "meshes" not in TF_base_functions.export_settings["filedata"]:
		TF_base_functions.export_settings["filedata"]["meshes"] = {}
	filedata = TF_base_functions.export_settings["filedata"]["meshes"]

	# create mesh
	mesh = mesh_obj.to_mesh(bpy.context.scene, True, 'PREVIEW')

	# get data from mesh
	model_type = mesh_obj.data.model_type
	if model_type == "%auto":
		model_type = TF_base_functions.export_settings["model_type"]
	elif model_type == "%inherit" and "parent" in mesh_object and "model_type" in mesh_object["parent"]:
		model_type = mesh_object["parent"]["model_type"]

	# get mesh dir
	mesh_dir = mesh_obj.data.mesh_dir
	if mesh_dir == "%auto":
		mesh_dir = TF_base_functions.export_settings["mesh_directory"]

	file_id = model_type + mesh_dir + "/"[:len(mesh_dir)] + mesh_obj.data.filename
	create_mesh = TF_base_functions.export_settings["subfiles"] or ("forceOverride" in mesh_object and mesh_object["forceOverride"])
	materials = []
	bounding_box = None

	# BB calculation
	if TF_base_functions.export_settings["bb"]:
		bounding_box = {"bbMax": [0.0, 0.0, 0.0], "bbMin": [0.0, 0.0, 0.0]}
		for i in range(8):
			for j in range(3):
				if mesh_obj.bound_box[i][j] < bounding_box["bbMin"][j]:
					bounding_box["bbMin"][j] = mesh_obj.bound_box[i][j]
				if mesh_obj.bound_box[i][j] > bounding_box["bbMax"][j]:
					bounding_box["bbMax"][j] = mesh_obj.bound_box[i][j]

	if file_id in filedata:
		# additional mesh with the same name, check materials only
		mesh_data = filedata[file_id]["data"]

		#if TF_base_functions.export_settings["game"] != "TPF2":
		for i in range(len(mesh_data["subMeshes"])):
			if "__binary" in mesh_data:
				material = mesh_obj.material_slots[mesh_data["__binary"]["mat_slots"][i]].material
				submesh = mesh_data["subMeshes"][i]
				if material:
					# get material
					mat_fileid = material_util.create_file_data({"object": material, "parent": {"object": mesh_obj, "model_type": model_type}}, "", errors)
					if TF_base_functions.export_settings["game"] != "TPF2" and mat_fileid not in submesh["materials"]:
						submesh["materials"].append(mat_fileid)
					materials.append(mat_fileid)
				else:
					errors.append({"level": 1, "message": "Materialslot %i in Mesh \"%s\" contains no material, this materialslot will be ignored!" %(i, mesh_object.data.name)})
	else:
		# first mesh
		if TF_base_functions.export_settings["game"] == "TPF2":
			mesh_data = {"subMeshes": [], "vertexAttr": {}, "__binary": {}}
		else:
			mesh_data = {"matConfigs": [], "subMeshes": [], "vertexAttr": {}, "__binary": {}}

		# get data from mesh
		mesh.calc_normals()
		world_mat = mesh_obj.matrix_world

		# check for ngons
		ngons_used = False
		for i in range(0, len(mesh.polygons)):
			ngons_used = ngons_used or (len(mesh.polygons[i].loop_indices) > 4)
		if ngons_used and create_mesh:
			if TF_base_functions.export_settings["ngons"]:
				for i in range(0, len(mesh_obj.data.polygons)):
					mesh_obj.data.polygons[i].select = (len(mesh_obj.data.polygons[i].loop_indices) > 4)
			errors.append({"level": 2, "message": "Mesh export failed: Mesh \"%s\" contains ngons!" %(mesh_obj.data.name)})
			return 0, None
		# no materials = no submeshes = no export ;)
		elif len(mesh.materials) <= 0 and create_mesh:
			errors.append({"level": 2, "message": "Mesh export failed: Mesh \"%s\" has no material!" %(mesh_obj.data.name)})
			return 0, None
		# uv layers required
		elif len(mesh.uv_layers) <= 0 and create_mesh:
			errors.append({"level": 2, "message": "Mesh export failed: Mesh \"%s\" has no UVMap!" %(mesh_obj.data.name)})
			return 0, None

		# create/update tangent data
		if create_mesh:
			mesh.calc_tangents(mesh.uv_layers[0].name)

		# base data structure
		vertex_struct = {
			"normal": {"count": 0, "offset": 0, "numComp": 3},
			"position": {"count": 0, "offset": 0, "numComp": 3},
			"tangent": {"count": 0, "offset": 0, "numComp": 4},
		}
		for i in range(len(mesh.uv_layers)):
			vertex_struct["uv"+str(i)] = {"count": 0, "offset": 0, "numComp": 2}

		# create data vars
		vertex = {}
		index_data = {i: {} for i in range(0, len(mesh.materials))}
		vertex_data = {}
		vertex_data_num = {}
		data_list = {}

		data = {
			"meshObj": mesh_obj,
			"mesh": mesh,
			"vertexStruct": vertex_struct,
		}
		TF_base_functions.call_plugin("create_mesh_data_storage", data, errors)

		if(TF_base_functions.export_settings["flipV"]):
			errors.append({"level": 1, "message": "Exporting with flipV = true"})
		else:
			errors.append({"level": 1, "message": "Exporting with flipV = false"})

		for key in vertex_struct.keys():
			# create key in storages if necessary
			if key not in data_list:
				data_list[key] = {}
			if key not in vertex_data:
				vertex_data[key] = []
			if key not in vertex_data_num:
				vertex_data_num[key] = 0
			for i in index_data.keys():
				if key not in index_data[i]:
					index_data[i][key] = []

		# read vertex data
		for i in range(0, len(mesh.polygons)):
			poly = mesh.polygons[i]
			uv_layers = [mesh.uv_layers[j] for j in range(len(mesh.uv_layers))]
			data["polygon"] = poly

			# check if poly can be exported
			if create_mesh and (TF_base_functions.export_settings["unselected_faces"] or poly.select) and (TF_base_functions.export_settings["hidden_faces"] or not poly.hide) and poly.area > 0:
				if len(poly.vertices) > 3:
					# triangulate quads
					vector_23 = (mesh.vertices[mesh.loops[poly.loop_indices[2]].vertex_index].co - mesh.vertices[mesh.loops[poly.loop_indices[3]].vertex_index].co).normalized().to_tuple(6)
					vector_30 = (mesh.vertices[mesh.loops[poly.loop_indices[3]].vertex_index].co - mesh.vertices[mesh.loops[poly.loop_indices[0]].vertex_index].co).normalized().to_tuple(6)
					if vector_23 == vector_30:
						poly_list = [[poly.loop_indices[0], poly.loop_indices[1], poly.loop_indices[3]], poly.loop_indices[1:4]]
					else:
						poly_list = [poly.loop_indices[0:3], [poly.loop_indices[2], poly.loop_indices[3], poly.loop_indices[0]]]
					uv_list = poly_list.copy()
				else:
					poly_list = [poly.loop_indices[0:3]]
					uv_list = [poly.loop_indices[0:3]]

				# get tris
				for j in range(0, len(poly_list)):
					for k in range(0, 3):
						vectors = {}
						vert = mesh.vertices[mesh.loops[poly_list[j][k]].vertex_index]

						# position
						vectors["position"] = vert.co

						# normal
						if poly.use_smooth:
							vectors["normal"] = vert.normal
						else:
							vectors["normal"] = poly.normal

						# uv
						for l in range(len(uv_layers)):
							vectors["uv"+str(l)] = uv_layers[l].data[uv_list[j][k]].uv
							if(TF_base_functions.export_settings["flipV"]):
								vectors["uv"+str(l)].y = 1 - vectors["uv"+str(l)].y

						# tangent
						tangent = mesh.loops[poly_list[j][k]].tangent
						vectors["tangent"] = Vector((tangent.x, tangent.y, tangent.z, mesh.loops[poly_list[j][k]].bitangent_sign))

						# call plugins, may support other vertex attributes
						data["vertex"] = vert
						data["vectors"] = vectors
						TF_base_functions.call_plugin("get_mesh_vertex_data", data, errors)

						# check if data already exists and store it, if not
						for key in vectors.keys():
							vec_string = vectors[key].to_tuple(6)

							if vec_string in data_list[key]:
								data_index = data_list[key][vec_string]
							else:
								data_index = vertex_data_num[key]
								vertex_data[key].append(vectors[key].copy())
								vertex_data_num[key] = vertex_data_num[key] + 1
								data_list[key][vec_string] = data_index
							index_data[poly.material_index][key].append(data_index)

						# garbage collection
						del vectors
			elif not create_mesh and poly.material_index in index_data:
				index_data[poly.material_index]["position"].append(0)

		del data

		# delete submeshes with zero faces
		for i in list(index_data.keys()):
			if "position" not in index_data[i] or len(index_data[i]["position"]) == 0:
				del index_data[i]
			else:
				# delete empty keys
				for key in list(index_data[i].keys()):
					if len(index_data[i][key]) == 0:
						del index_data[i][key]
		
		# read materials
		submeshes = []
		mat_slots = []
		for i in range(0, len(mesh_obj.material_slots)):
			material = mesh_obj.material_slots[i].material
			if material:
				if i in index_data:
					if TF_base_functions.export_settings["game"] != "TPF2":
						submesh = {"indices": {}, "materials": []}
					else:
						submesh = {"indices": {} }
					for key in index_data[i].keys():
						submesh["indices"][key] = {'count': 0, 'offset': 0}

					# get material
					mat_fileid = material_util.create_file_data({"object": material, "parent": {"object": mesh_obj, "model_type": model_type}}, "", errors)
					if TF_base_functions.export_settings["game"] != "TPF2":
						submesh["materials"].append(mat_fileid)
					materials.append(mat_fileid)
					
					submeshes.append(submesh)
					mat_slots.append(i)
			else:
				errors.append({"level": 1, "message": "Materialslot %i in Mesh \"%s\" contains no material, this materialslot will be ignored!" %(i, mesh_object.data.name)})

		# calculate length of binary data
		for key in vertex_data.keys():
			if len(vertex_data[key]) > 0:
				for i in range(0, len(submeshes)):
					if key in submeshes[i]["indices"]:
						submeshes[i]["indices"][key]["count"] = 4*len(index_data[mat_slots[i]][key])
				if key not in vertex:
					vertex[key] = vertex_struct[key]
				vertex[key]["count"] = 4 * vertex[key]["numComp"] * len(vertex_data[key])

		# calculate offsets
		key_list = []
		offset = 0
		for key in vertex.keys():
			key_list.append(key)
			vertex[key]["offset"] = offset
			offset = offset + vertex[key]["count"]
		for i in range(0, len(submeshes)):
			for key in submeshes[i]["indices"].keys():
				submeshes[i]["indices"][key]["offset"] = offset
				offset = offset + submeshes[i]["indices"][key]["count"]

		# store data
		mesh_data["subMeshes"] = submeshes
		mesh_data["vertexAttr"] = vertex
		if TF_base_functions.export_settings["game"] != "TPF2":
			mesh_data["matConfigs"] = []
		mesh_data["__binary"] = {"index_data": index_data, "vertex_data": vertex_data, "mat_slots": mat_slots}
		mesh_data["__key_list"] = key_list

		# call plugins
		data = {
			"meshObj": mesh_obj,
			"meshData": mesh_data,
			"fileID": file_id,
			"modelType": model_type,
		}
		TF_base_functions.call_plugin("save_mesh_data_once", data, errors)
		del data

		# create default matConfig if none exists
		if TF_base_functions.export_settings["game"] != "TPF2" and len(mesh_data["matConfigs"]) == 0:
			mesh_data["matConfigs"].append([0 for i in range(len(submeshes))])

		# save data for export
		filedata[file_id] = {"data": mesh_data, "model_type": model_type+mesh_dir + "/"[:len(mesh_dir)], "name": mesh_name, "create": create_mesh}
		if "forceOverride" in mesh_object:
			filedata[file_id]["forceOverride"] = mesh_object["forceOverride"]

	# execute for every mesh object
	# call plugins
	data = {
		"meshObj": mesh_obj,
		"meshData": mesh_data,
		"fileID": file_id,
		"modelType": model_type,
	}
	TF_base_functions.call_plugin("save_mesh_data_always", data, errors)

	if TF_base_functions.export_settings["game"] != "TPF2":
		matConfig = []
		if len(materials) == len(mesh_data["subMeshes"]):
			for i in range(len(mesh_data["subMeshes"])):
				if materials[i] in mesh_data["subMeshes"][i]["materials"]:
					matConfig.append(mesh_data["subMeshes"][i]["materials"].index(materials[i]))
				else:
					matConfig.append(0)

		if len(mesh_data["matConfigs"]) == 0:
			matConfigID = len(mesh_data["matConfigs"])
			if matConfig != [] and matConfig in mesh_data["matConfigs"]:
				matConfigID = mesh_data["matConfigs"].index(matConfig)
			else:
				mesh_data["matConfigs"].append(matConfig)
	
	del data
	
	# delete created mesh
	bpy.data.meshes.remove(mesh)
	del mesh

	if TF_base_functions.export_settings["game"] == "TPF2":
		return materials, bounding_box
	else:
		return matConfigID, bounding_box

def write_file(file, errors):
	filename = file["name"] + ".msh"
	filepath = file["model_type"]

	id = filepath + filename

	resfilepath = "models/mesh/" + filepath
	resfilename = resfilepath + filename

	if TF_base_functions.export_settings["dir_structure"]:
		osfilepath = TF_base_functions.export_settings["basepath"]+resfilepath
	else:
		osfilepath = TF_base_functions.export_settings["basepath"]

	osfilename = osfilepath + filename

	mesh_data = file["data"]
	if (not writer.isfile(osfilename)) or TF_base_functions.export_settings["override"] or ("forceOverride" in file and file["forceOverride"]):
		if "__binary" in mesh_data and "__key_list" in mesh_data:
			if not writer.exists(osfilepath):
				writer.makedirs(osfilepath)

			output = open(osfilename + ".blob", "wb")
			for key in mesh_data["__key_list"]:
				for i in range(len(mesh_data["__binary"]["vertex_data"][key])):
					for j in range(len(mesh_data["__binary"]["vertex_data"][key][i])):
						output.write(struct.pack("<f", mesh_data["__binary"]["vertex_data"][key][i][j]))
			for j in sorted(mesh_data["__binary"]["index_data"].keys()):
				for key in mesh_data["__key_list"]:
					for i in range(len(mesh_data["__binary"]["index_data"][j][key])):
						output.write(struct.pack("<L", mesh_data["__binary"]["index_data"][j][key][i]))
			output.close()

	if (not writer.isfile(osfilename)) or TF_base_functions.export_settings["override"] or ("forceOverride" in file and file["forceOverride"]):
		if not writer.exists(osfilepath):
			writer.makedirs(osfilepath)

		filetext = ""
		# additional code, to do: not restricted to require
		if "__requires" in file["data"]:
			for req in file["data"]["__requires"]:
				filetext += req + '\n'

		filetext += "function data()\nreturn "
		filetext += writer.write_var_struct_string(file["data"], "", 0, True, False)
		filetext += "end"

		data = {
			"text": filetext,
			"meshData": mesh_data,
			"fileID": id,
		}
		TF_base_functions.call_plugin("write_mesh_file", data, errors)

		output = open(osfilename, 'w', encoding="utf-8")
		output.write(data["text"])
		output.close()

	TF_base_functions.export_settings["filelist"].append([osfilename, "res/" + resfilename])

# Classes
class TFMeshSetting(bpy.types.PropertyGroup):
	@classmethod
	def register(cls):
		bpy.types.Mesh.tf_mesh = bpy.props.PointerProperty(type=cls)

	@classmethod
	def unregister(cls):
		del bpy.types.Mesh.tf_mesh

class TFMeshSettingsPanel(bpy.types.Panel):
	bl_label = "Train Fever Mesh"
	bl_space_type = "PROPERTIES"
	bl_region_type = "WINDOW"
	bl_context = "data"

	@classmethod
	def poll(cls, context):
		if (context and context.object and context.object.type in {'MESH'}):
			return True
		return False

	def draw(self, context):
		if context.object.type in {'MESH'}:
			self.draw_mesh(context, self.layout)

	def draw_mesh(self, context, layout):
		mydata = context.object.data

		if mydata:
			TF_base_functions.call_plugin("draw_mesh_settings_panel", {"layout": layout, "mydata": mydata}, [])

			TF_base_functions.draw_internal_box(mydata, layout, "mesh")

__class_list = [
	TFMeshSetting,
	TFMeshSettingsPanel,
]

def register():
	for class_type in __class_list:
		bpy.utils.register_class(class_type)

	# general props: model_type and mesh_dir
	bpy.types.Mesh.model_type = bpy.props.StringProperty(default="%auto")
	bpy.types.Mesh.filename = bpy.props.StringProperty(default="%auto")
	bpy.types.Mesh.mesh_dir = bpy.props.StringProperty(default="%auto")
	
	TF_base_functions.export_file_ext["msh"] = (("msh", "Mesh (.msh)", "", 1))

def unregister():
	# general props: model_type and mesh_dir
	del bpy.types.Mesh.model_type
	del bpy.types.Mesh.filename
	del bpy.types.Mesh.mesh_dir
	
	del TF_base_functions.export_file_ext["msh"]

	for i in range(len(__class_list)-1, -1, -1):
		bpy.utils.unregister_class(__class_list[i])